/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.cable.cfw;

import java.net.URL;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;
import javax.swing.SwingUtilities;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;
import java.util.logging.Logger;
import java.util.logging.Level;

// XAL framework stuff
import gov.sns.application.*;

// general app utilities
import gov.anl.aps.irmis.apps.SwingThreadWork;
import gov.anl.aps.irmis.apps.QueuedExecutorUtil;
import gov.anl.aps.irmis.apps.ProgressDialog;
import gov.anl.aps.irmis.login.LoginUtil;

// other IRMIS sub-applications
import gov.anl.aps.irmis.apps.component.cfw.ComponentDocument;

// persistence layer
import gov.anl.aps.irmis.persistence.component.Component;
import gov.anl.aps.irmis.persistence.component.ComponentRelationship;
import gov.anl.aps.irmis.persistence.component.ComponentRelationshipType;
import gov.anl.aps.irmis.persistence.component.ComponentPort;
import gov.anl.aps.irmis.persistence.component.ComponentType;
import gov.anl.aps.irmis.persistence.component.Cable;
import gov.anl.aps.irmis.persistence.audit.AuditAction;
import gov.anl.aps.irmis.persistence.audit.AuditActionType;
import gov.anl.aps.irmis.persistence.DAOException;

// service layer
import gov.anl.aps.irmis.service.IRMISException;
import gov.anl.aps.irmis.service.shared.AuditService;
import gov.anl.aps.irmis.service.pv.PVService;
import gov.anl.aps.irmis.service.component.ComponentTypeService;
import gov.anl.aps.irmis.service.component.ComponentService;


/**
 * The primary controller class for IRMIS Cable Viewer application.
 * Provides "action" methods to to query the IRMIS database for cable/conductor
 * related data. These action methods are typically invoked by the 
 * <code>CableWindow</code>.
 * Any time-consuming activity should be queued up using the 
 * <code>queuedExecutor</code>, which is similar to the SwingWorker concept.
 */
public class CableDocument extends XalInternalDocument {

	/** The main data model of this application */
	private CableModel _model;

    // Hold reference to parent desktop frame. Used to put up busy icon during work.
    JFrame _appFrame;

	/** Create a new empty document */
    public CableDocument() {
        this(null);
        _model = new CableModel();
    }

	/** Create a new document from url */
    public CableDocument(final URL url) {
        //setSource(url);
        _model = new CableModel();
       _appFrame = ((DesktopApplication)Application.getApp()).getDesktopFrame();
        // don't have anything here yet
    }

    public void saveDocumentAs(final URL url) {
        return;
    }

    /**
     * Specifies the document file suffixes supported by this document.
     * @return An array of file suffixes corresponding to writable files
     */
    public String[] writableDocumentTypes() {
        return new String[] {};
	}

    /**
	 * Overriding default window title to my liking.
     * @param newTitle The new title for this document.
     */
    public void setTitle( final String newTitle ) {
        String modifiedTitle = "idt::cable" + newTitle;
		super.setTitle( modifiedTitle );
        _documentListenerProxy.titleChanged( this, title );
    }	
            
    /**
     * Make a main window by instantiating the my custom window.
     */
    public void makeMainWindow() {
        _mainWindow = new CableWindow(this);

        // Initialize application data from database and do a full redraw of GUI.
        actionReload();
        actionRedraw();
    }

    /**
     * Clear out all the data in the model.
     */
    public void actionResetModel() {
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                public void doNonUILogic() {
                    _model.reset();
                }
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_HIERARCHIES));
                }
            });
    }

    /**
     * Queue up job for background thread to gather the current housing and
     * control hierarchies, along with any other data we will need
     * regularly during execution. When complete, fire off 
     * <code>CableModelEvent</code> to notify GUI that new data is available
     * for display.
     */
    public void actionReload() {
        
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                private Component siteComponent = null;
                private List roomComponents = null;
                private Component networkComponent = null;
                private Component utilityComponent = null;
                private List componentTypes = null;
                private List systemList = null;
                public void doNonUILogic() {

                    // Get current set of Components of type room, and put into model
                    _model.setIRMISException(null);
                    try {
                        // put up progress dialog, since this can take a while
                        ProgressDialog.show(_appFrame,1,4);

                        // housing hierarchy
                        ComponentType siteCT = 
                            ComponentTypeService.findComponentTypeByName("Site");
                        List sites = ComponentService.findComponentsByType(siteCT);
                        // there better be just one
                        if (sites.size() != 1) {
                            throw new IRMISException("Unable to find root of housing hierarchy (Component Type: Site)");
                        }
                        siteComponent = (Component)sites.get(0);
                        // get all 2nd level housing components (rooms) into one list
                        List childRels = 
                            siteComponent.getChildRelationships(ComponentRelationshipType.HOUSING);
                        roomComponents = new ArrayList();
                        if (childRels != null) {
                            Iterator childIt = childRels.iterator();
                            while (childIt.hasNext()) {
                                ComponentRelationship rel = (ComponentRelationship)childIt.next();
                                Component buildingComponent = rel.getChildComponent();
                                if (buildingComponent.getComponentType().getComponentTypeName().equals("Room")) {
                                    roomComponents.add(buildingComponent); // check for rooms at first level
                                } else {
                                    List roomList = 
                                        buildingComponent.getChildRelationships(ComponentRelationshipType.HOUSING);
                                    if (roomList != null && roomList.size() > 0) {
                                        Iterator roomIt = roomList.iterator();
                                        while (roomIt.hasNext()) {
                                            ComponentRelationship rr = (ComponentRelationship)roomIt.next();
                                            Component room = rr.getChildComponent();
                                            if (room.getComponentType().getComponentTypeName().equals("Room"))
                                                roomComponents.add(room);
                                        }
                                    }
                                }
                            }
                        }

                        ProgressDialog.setValue(1);

                        // control hierarchy
                        ComponentType networkCT = 
                            ComponentTypeService.findComponentTypeByName("Network");
                        List networks = ComponentService.findComponentsByType(networkCT);
                        // there better be just one
                        if (networks.size() != 1) {
                            throw new IRMISException("Unable to find root of control hierarchy (Component Type: Network)");
                        }
                        networkComponent = (Component)networks.get(0);
                        ProgressDialog.setValue(2);

                        // get this info, since we need it many times
                        componentTypes = ComponentTypeService.findComponentTypeList();
                        ProgressDialog.setValue(3);

                        // also, bring in lazy collection we'll need
                        Iterator it = componentTypes.iterator();
                        while (it.hasNext()) {
                            ComponentType ct = (ComponentType)it.next();
                            ct.getComponentTypeInterfaces().iterator();
                        }

                        // get list of ioc system names
                        systemList = PVService.findSystems();

                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);

                    } finally {
                        ProgressDialog.destroy();
                    }
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.setSiteComponent(siteComponent);
                    _model.setRoomComponents(roomComponents);
                    _model.setNetworkComponent(networkComponent);
                    _model.setComponentTypes(componentTypes);
                    _model.setSystems(systemList);
                    //_model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_HIERARCHIES));
                }
            });
    }

    public void actionRedraw() {
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                public void doNonUILogic() {
                }
                
                public void doUIUpdateLogic() {
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_HIERARCHIES));
                }
            });
    }


    public void actionSetComponentBUsingCable(Cable cable) {
        final Cable selectedCable = cable;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                ComponentPort port = null;
                public void doNonUILogic() {
                    // cable can attach at either end, so figure out which is which
                    ComponentPort portA = selectedCable.getComponentPortA();
                    ComponentPort portB = selectedCable.getComponentPortB();
                    if (portA != null && portA.equals(_model.getSelectedPort1()))
                        port = portB;
                    else
                        port = portA;
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_COMPONENT_B_AND_PORT, port));
                }
            });        
    }

    public void actionAddNewCable() {
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                Cable newCable = null;
                public void doNonUILogic() {
                    try {
                        _model.setIRMISException(null);
                        newCable = new Cable();
                        newCable.setLabel("");
                        newCable.setColor("");
                        newCable.setDestinationDescription("");
                        newCable.setVirtual(false);
                        _model.getSelectedPort1().addCableA(newCable);
                        if (_model.getSelectedPort2() != null)  // don't do this if single-ended
                            _model.getSelectedPort2().addCableB(newCable);


                        AuditAction auditAction = 
                            AuditService.createAuditAction(AuditActionType.ADD_CABLE,
                                                           "new cable", LoginUtil.getUsername(),
                                                           newCable.getId());                        
                        _model.addAction(auditAction);
                        
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);
                    }
                    requestResetOfDependentDocuments();
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.CABLE_ADDED, newCable));
                }
            });        
    }

    public void actionUpdateCable(Cable cable, String updateDescription) {
        final Cable selectedCable = cable;
        final String updateDesc = updateDescription;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                public void doNonUILogic() {
                    try {
                        _model.setIRMISException(null);
                        AuditAction auditAction = 
                            AuditService.createAuditAction(AuditActionType.EDIT_CABLE,
                                                           updateDesc, LoginUtil.getUsername(),
                                                           selectedCable.getId());
                        _model.addAction(auditAction);
                        
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                    }

                    requestResetOfDependentDocuments();
                }
                
                public void doUIUpdateLogic() {
                    // do nothing
                }
            });        
    }

    public void actionSaveCables(List cables) {
        final List cableList = cables;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                public void doNonUILogic() {
                    try {
                        _model.setIRMISException(null);
                        Iterator cableIt = cableList.iterator();
                        while (cableIt.hasNext()) {
                            Cable cbl = (Cable)cableIt.next();
                            ComponentService.saveCable(cbl);
                        }

                        List actionList = _model.getActionList();
                        Iterator actionIt = actionList.iterator();
                        while (actionIt.hasNext()) {
                            AuditAction aa = (AuditAction)actionIt.next();
                            AuditService.saveAuditAction(aa);
                        }
                        _model.clearActionList();
                        
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                    }
                    requestResetOfDependentDocuments();
                }
                
                public void doUIUpdateLogic() {
                    // do nothing
                }
            });        
    }

    public void actionRemoveCable(Cable cable) {
        final Cable selectedCable = cable;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                public void doNonUILogic() {
                    try {
                        _model.setIRMISException(null);
                        selectedCable.setMarkForDelete(true);
                        ComponentPort p1 = selectedCable.getComponentPortA();
                        ComponentPort p2 = selectedCable.getComponentPortB();
                        if (p1 != null) {
                            List cablesA = p1.getCablesA();
                            cablesA.remove(selectedCable);
                        }
                        if (p2 != null) {
                            List cablesB = p2.getCablesB();
                            cablesB.remove(selectedCable);
                        }
                        ComponentService.saveCable(selectedCable);
                        AuditAction auditAction = 
                            AuditService.createAuditAction(AuditActionType.REMOVE_CABLE,
                                                           "remove cable", LoginUtil.getUsername(),
                                                           selectedCable.getId());
                        AuditService.saveAuditAction(auditAction);
                        //_model.addAction(auditAction);
                        
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);
                    }
                    requestResetOfDependentDocuments();
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.CABLE_REMOVED, selectedCable));
                }
            });        
    }

    /**
     * Query db to find the cable between port1 and port2 (if any).
     */
    public void actionFindCable(ComponentPort port1, ComponentPort port2) {
        final ComponentPort selectedPort1 = port1;
        final ComponentPort selectedPort2 = port2;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                List cables = null;
                Cable cable = null;
                public void doNonUILogic() {
                    _model.setIRMISException(null);                    
                    try {
                        // evict any prior cables from cache
                        List previousCables = _model.getCables();
                        ComponentService.clearCache(previousCables);

                        // find new cables 
                        cables = ComponentService.findCablesByPort(selectedPort1);
                        if (cables != null) {
                            Iterator cableIt = cables.iterator();
                            while (cableIt.hasNext()) {
                                Cable tempCable = (Cable)cableIt.next();
                                ComponentPort portA = tempCable.getComponentPortA();
                                ComponentPort portB = tempCable.getComponentPortB();
                                if ((portA != null && portA.equals(selectedPort2)) ||
                                    (portB != null && portB.equals(selectedPort2))) {
                                    cable = tempCable;
                                    break;
                                }
                            }
                        }
                        _model.setSelectedCable(cable);
                        
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);
                    }
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                        _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_CABLE,cable));
                }
            });

    }

    /**
     * Query db to find cable(s) attached to given port (if any).
     */
    public void actionFindCables(ComponentPort port) {
        final ComponentPort selectedPort = port;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                List cables = null;
                public void doNonUILogic() {
                    _model.setIRMISException(null);                    
                    try {
                        // evict any prior cables from cache
                        List previousCables = _model.getCables();
                        ComponentService.clearCache(previousCables);
                        // get new list of cables
                        cables = ComponentService.findCablesByPort(selectedPort);
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);
                    }
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                        _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_CABLES,cables));
                }
            });

    }

    /**
     * Query db to find cable(s) attached to given port (if any).
     */
    public void actionFindCablesByLabel(String labelSearch) {
        final String labelSearchText = labelSearch;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {
                List cables = null;
                public void doNonUILogic() {
                    _model.setIRMISException(null);                    
                    try {
                        // evict any prior cables from cache
                        List previousCables = _model.getCables();
                        ComponentService.clearCache(previousCables);
                        // get new list of cables
                        cables = ComponentService.findCablesByLabel(labelSearchText);
                    } catch (IRMISException ie) {
                        ie.printStackTrace();
                        _model.setIRMISException(ie);
                    }
                }
                
                public void doUIUpdateLogic() {
                    // fire off notification of modification to listeners
                        _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.NEW_CABLES,cables));
                }
            });

    }

    /**
     * Load up all children of given component c in a particular hierarchy. 
     * Normally these children are lazily loaded, so this is explicit way 
     * to get it done all at once.
     */
    public void actionFindAllChildren(Component c, int hierarchy) {
        final Component selectComponent = c;
        final int h = hierarchy;
        QueuedExecutorUtil.execute(new SwingThreadWork(_appFrame) {

                public void doNonUILogic() {
                    _model.setIRMISException(null);
                    // ignore returned list since we are just making sure
                    // object cache is pre-loaded
                    ComponentService.findAllChildren(selectComponent, h);
                }

                public void doUIUpdateLogic() {
                    // add this selection request to Swing event queue
                    _model.notifyCableModelListeners(new CableModelEvent(CableModelEvent.EXPAND_CHILDREN, selectComponent));
                }
            });

    }

    /**
     * Goes through IRMIS desktop list of documents and does an actionReset()
     * on those that are dependent on our state.
     */
    private void requestResetOfDependentDocuments() {
        Application app = Application.getApp();
        List docs = app.getDocuments();
        
        /*
        // iterate, looking for all other idt::cable docs
        Iterator docIt = docs.iterator();
        CableDocument cDoc = null;
        while (docIt.hasNext()) {
            XalInternalDocument doc = (XalInternalDocument)docIt.next();

            // If doc is an instance of idt::cable (and not us), request reset
            if (doc.getTitle().equals("idt::cable") && !doc.equals(this)) {
                cDoc = (CableDocument)doc;
                cDoc.actionReset();
            }
        }
        */
    }

	/**
	 * Get the launch model which represents the main model of this document
	 * @return the main model of this document
	 */
	public CableModel getModel() {
		return _model;
	}

    private CableWindow myWindow() {
        return (CableWindow)_mainWindow;
    }
}
