/*
   Copyright (c) 2004-2005 The University of Chicago, as Operator
   of Argonne National Laboratory.
*/
package gov.anl.aps.irmis.apps.cable.cfw;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

// other IRMIS sub-applications
import gov.anl.aps.irmis.apps.component.cfw.ComponentDocument;

// persistence layer
import gov.anl.aps.irmis.persistence.component.Component;
import gov.anl.aps.irmis.persistence.component.ComponentPort;
import gov.anl.aps.irmis.persistence.component.ComponentType;
import gov.anl.aps.irmis.persistence.component.Cable;
import gov.anl.aps.irmis.persistence.component.ComponentRelationshipType;
import gov.anl.aps.irmis.persistence.audit.AuditAction;

// service layer
import gov.anl.aps.irmis.service.IRMISException;
import gov.anl.aps.irmis.service.pv.PVSearchParameters;

/**
 * Application data model for IRMIS Cable Viewer. 
 * A list of listeners interested in this data is maintained, and
 * notified as requested.
 */
public class CableModel {

    // list of others interested in changes to the data model
    private List cableModelListenersList = new ArrayList();

    // used to flag a problem talking to db
    private IRMISException ie = null;

    private ComponentPort selectedPort1 = null;
    private ComponentPort selectedPort2 = null;

    private List cables = new ArrayList();
    private Cable selectedCable = null;

    private Component siteComponent = null;  // housing root
    private List roomComponents = null;      // all 1st level housing components (rooms)
    private Component networkComponent = null;   // control root

    private Component selectedComponent1 = null; // selection in hierarchy
    private Component selectedComponent2 = null; // selection in hierarchy

    private List actionList = new ArrayList();  // list of ConfigureAction

    private List componentTypes = null; // list of possible types
    private List filteredHousingComponentTypes = null; // same list after constraints applied
    private List filteredControlComponentTypes = null; // same list after constraints applied

    private List systems = null; // list of unique ioc system names
    // end from ComponentModel

	/**
	 * Do nothing constructor
	 */
	public CableModel() {
	}

    public void reset() {
        siteComponent = null;
        networkComponent = null;
        selectedComponent1 = null;
        selectedComponent2 = null;
        componentTypes = null;
        filteredHousingComponentTypes = null;
        filteredControlComponentTypes = null;
        selectedPort1 = null;
        selectedPort2 = null;
        cables = new ArrayList();
        actionList = new ArrayList();
        selectedCable = null;
    }

    public ComponentPort getSelectedPort1() {
        return this.selectedPort1;
    }
    public void setSelectedPort1(ComponentPort value) {
        this.selectedPort1 = value;
    }

    public ComponentPort getSelectedPort2() {
        return this.selectedPort2;
    }
    public void setSelectedPort2(ComponentPort value) {
        this.selectedPort2 = value;
    }

    public List getActionList() {
        return actionList;
    }
    public void addAction(AuditAction aa) {
        actionList.add(aa);
    }
    public void clearActionList() {
        actionList.clear();
    }

    public List getCables() {
        return this.cables;
    }
    public void setCables(List value) {
        this.cables = value;
    }

    public Cable getSelectedCable() {
        return this.selectedCable;
    }
    public void setSelectedCable(Cable value) {
        this.selectedCable = value;
    }

    public void addCableModelListener(CableModelListener l) {
        cableModelListenersList.add(l);
    }

    public void notifyCableModelListeners() {
        notifyCableModelListeners(null);
    }



    // begin from ComponentModel

    public Component getSiteComponent() {
        return siteComponent;
    }
    public void setSiteComponent(Component value) {
        siteComponent = value;
    }

    public List getRoomComponents() {
        return roomComponents;
    }
    public void setRoomComponents(List value) {
        roomComponents = value;
    }

    public Component getNetworkComponent() {
        return networkComponent;
    }
    public void setNetworkComponent(Component value) {
        networkComponent = value;
    }

    public Component getSelectedComponent1() {
        return selectedComponent1;
    }
    public void setSelectedComponent1(Component value) {
        selectedComponent1 = value;
    }

    public Component getSelectedComponent2() {
        return selectedComponent2;
    }
    public void setSelectedComponent2(Component value) {
        selectedComponent2 = value;
    }

    public List getComponentTypes() {
        return componentTypes;
    }
    public void setComponentTypes(List value) {
        componentTypes = value;
    }

    public List getSystems() {
        return systems;
    }
    public void setSystems(List value) {
        systems = value;
    }

    public List getFilteredComponentTypes(int relationshipType) {
        switch (relationshipType) {
        case ComponentRelationshipType.CONTROL: {
            return filteredControlComponentTypes;
        }
        case ComponentRelationshipType.HOUSING: {
            return filteredHousingComponentTypes;
        }
        default: {
            return null;
        }
        }
    }
    public void setFilteredComponentTypes(List value, int relationshipType) {
        switch (relationshipType) {
        case ComponentRelationshipType.CONTROL: {
            filteredControlComponentTypes = value;
            break;
        }
        case ComponentRelationshipType.HOUSING: {
            filteredHousingComponentTypes = value;
            break;
        }
        }
    }
    // end from ComponentModel


    /**
     * Notify registered listeners of the given <code>CableModelEvent</code>.
     *
     * @param modelEvent encapsulates one of several possible events
     */
    public void notifyCableModelListeners(CableModelEvent modelEvent) {
        Iterator it = cableModelListenersList.iterator();
        while (it.hasNext()) {
            CableModelListener l = (CableModelListener)it.next();
            l.modified(modelEvent);
        }
    }

    public IRMISException getIRMISException() {
        return this.ie;
    }
    public void setIRMISException(IRMISException ie) {
        this.ie = ie;
    }

}
