package NetworkCrawlerDBLayer;

# This package provides routines for writing ioc interface network 
# data from accel_sw file to a MySQL database. 

use DBI qw(:sql_types);
require Exporter;
use LogUtil;
use Blowfish_PP;
use File::stat;

our @ISA = qw(Exporter);
our @EXPORT = qw( &set_test_mode &set_home_dir &set_decryption_key
                  &zero_out_network_data &load_new_network_data );
our @EXPORT_OK = qw();
our $VERSION = 1.00;

my $test_mode = 0;  # if 1, do no db access
my $home_dir = "."; # dir from which parent crawler is run
my $db_properties_path = "db.properties";
my $dbh;
my $db_vendor; # established on first db_connect
my $decryption_key; # optional 16 byte blowfish key
my $update_table="aps_ioc";

#=============================================================================
# Public
sub set_test_mode {
    $test_mode = $_[0];
}

#=============================================================================
# Public
# Sets the directory where db.properties file can be found. 
sub set_home_dir {
    $home_dir = $_[0];

    $db_properties_path = $home_dir . "/db.properties";
}

#=============================================================================
# Public
# Sets the 16 char blowfish decryption key. If this is defined, this module
# will assume the db.properties file is encrypted with this key.
sub set_decryption_key {
    $decryption_key = $_[0];
}

#=============================================================================
sub db_connect {

    #connect to database
    my $dsn;
    my $dbh;

    # Look up connection parameters in external db.properties file.
    # If a decryption key has been supplied, use blowfish algorithm
    # to decrypt the contents of the file.
    
    my $db_host;
    my $db_database;
    my $db_user;
    my $db_password;

    if (!-e $db_properties_path) {
        LogUtil::log('error',"Unable to find db.properties file");
        LogUtil::log('error',"Exiting...");
        exit;
    }

    # Read in contents of db.properties (decrypting if key defined)
    my $blowfish_object;
    if ($decryption_key) {
        my $packed_key = pack("a16",$decryption_key);
        $blowfish_object = Crypt::Blowfish_PP->new($packed_key);
        if (!$blowfish_object) {
            LogUtil::log('error',"The given blowfish key appears to be invalid");
            LogUtil::log('error',"Exiting...");
            exit;
        }
    }
    open(DBPROPS,$db_properties_path);
    my $inbuf;
    my $input_text;
    while (my $read_len = read(DBPROPS, $inbuf, 8)) {
        if ($blowfish_object) {
            my $decrypted_text = $blowfish_object->decrypt($inbuf);
            $input_text = $input_text . $decrypted_text;
        } else {
            $input_text = $input_text . $inbuf;
        }
    }
    close(DBPROPS);

    # Parse up the db.properties text
    my $props_line;
    while ($input_text =~ /^([^\n]*)\n(.*)/s) {
        $props_line = $1;
        $input_text = $2;
        if ($props_line =~ /(.*)=(.*)/) {
            $prop_name = $1;
            $prop_val = $2;
            if ($prop_name eq "db.host") {
                $db_host = $prop_val;
            } elsif ($prop_name eq "db.database") {
                $db_database = $prop_val;
            } elsif ($prop_name eq "db.user") {
                $db_user = $prop_val;
            } elsif ($prop_name eq "db.password") {
                $db_password = $prop_val;
            } elsif ($prop_name eq "db.vendor") {
                # db_vendor is package global
                $db_vendor = $prop_val;
            }
        }
    }
    if (!$db_host || !$db_database || !$db_user || !$db_password ||!$db_vendor) {
        LogUtil::log('error',"Unable to read db connection parameters from db.properties");
        LogUtil::log('error',"Exiting...");
        exit;
    }


    eval {
        $dsn =  "DBI:$db_vendor:$db_database:$db_host";

        $dbh = DBI->connect($dsn,$db_user,$db_password,
                            {AutoCommit=>0, RaiseError=>1, PrintError=>0, 
                             FetchHashKeyName=>'NAME_lc'});
    };
    if ($@) {
        LogUtil::log('error',"Unable to connect to db $@");
        LogUtil::log('error',"Exiting...");
        exit;
    }
    return $dbh;
}

#=============================================================================
# Public
sub zero_out_network_data {
  my $ioc_id = $_[0];

  my $update_line;
  my $update;
  my $sth;

  if ($test_mode) {
    LogUtil::log('info',"Test mode. No database changes made.");
    $emailBody .= "Test mode. Nodatabase changes made.\n";
    return 0;
  }

# Update data for the specified ioc_id 
  $update_line = "UPDATE $update_table ";
  $update_line .= "set ";
  $update_line .= "PrimEnetSwitch='', ";
  $update_line .= "PrimEnetBlade='', ";
  $update_line .= "PrimEnetPort='0', ";
  $update_line .= "SecEnetSwitch='', ";
  $update_line .= "SecEnetBlade='', ";
  $update_line .= "SecEnetPort='0' ";

  $update = qq($update_line);

  if (!$dbh) {
    $dbh = DBLayer::db_connect();
    if (!$dbh) {
      LogUtil::log('error',"Cannot connect to database");
      $email_body .= "ERROR: Cannot connect to database.\n";
      return -1;
    }
  }

  eval {
    $sth = $dbh->prepare($update);
    $sth->execute();
    $sth->finish();

#    $dbh->commit();
  };
  if($@) {
    LogUtil::log('error',"Zeroing out network data: $@");   
    $email_body .= "ERROR: Cannot zero out network data.\n";
    return -1;
  }

  LogUtil::log('info',"Zeroed out network data in table $update_table.");   
  $email_body .= "Zeroed out network data in table $update_table.\n";
  return 0;
}

#=============================================================================
# Public
sub load_new_network_data {
  my $infile = $_[0];
  my $network_hits_ref = $_[1];

  my $update_count = 0;

  LogUtil::log('info',"Updating all ioc network data.");   
  $email_body .= "Updating all ioc network data.\n";

  if ($test_mode) {
    LogUtil::log('info',"Test mode. No database changes made.");
    $emailBody .= "Test mode. Nodatabase changes made.\n";
    return 0;
  }

  if (!$dbh) {
    LogUtil::log('error',"Cannot connect to database");
    $email_body .= "ERROR: Cannot connect to database.\n";
    return -1;
  }

  # Loop over array items, updating network data in database
  foreach $row (@$network_hits_ref) {

    return -1 if update_database_record($row) == -1;   
    $update_count++;
  }

  LogUtil::log('info',"Number of database updates = $update_count.");
  $email_body .= "\nNumber of database updates = $update_count\n";

  $dbh->commit();

  return 0;
}

#=============================================================================
sub update_database_record {
  my ($ioc,$Etype,$switch,$blade,$port)= @_;   

  my $ioc_id = 0;
  my $ioc_row;
  my $select_line;
  my $update_line;
  my $update_request;
  my $sth;

  if ( $Etype =~ m/p/ ) {
    $msg = sprintf("%-16s Primary Ethernet:   %10s:%3s:%-2s",
       $ioc,$switch,$blade,$port);
  }
  if ( $Etype =~ m/s/ )  {
    $msg=sprintf("%-16s Secondary Ethernet: %10s:%3s:%-2s",
       $ioc,$switch,$blade,$port);
  }
  LogUtil::log('info',"Update Data: $msg");   
  $email_body .= "Update data: $msg\n";

  # Get ioc_id from database
  $select_line = "select ioc_id from ioc where ioc_nm='$ioc'";
  eval {
    $sth = $dbh->prepare($select_line);
    $sth->execute();
  };
  if($@) {
    LogUtil::log('error',"database select error for $ioc: $@");   
    $email_body .= "ERROR: select error for $ioc: $@ \n";
    return -1;
  }
  if ($ioc_row = $sth->fetchrow_hashref) {
    $ioc_id = $ioc_row->{ioc_id};
  }
  if ($ioc_id == 0) {
    # This is not fatal. May be a bad record in new switch data.
    LogUtil::log('warning',
          "Update failed: cannot find ioc $ioc in database");   
    $email_body .= "Update failed: cannot find ioc $ioc in database\n";
    return 0;
  }

  #printf ( "%-20s %s %10s:%-3s:%-2s\n",$ioc,$Etype,$switch,$blade,$port); 
        
  # Update data for the specified ioc_id 
  $update_line = "UPDATE $update_table ";
  $update_line .= "set ";

  if ( $Etype =~ m/p/ ) {
    $msg = sprintf("%-16s Primary Ethernet: %10s:%3s:%-2s",
       $ioc,$switch,$blade,$port);
    $update_line .= "PrimEnetSwitch='$switch', ";
    $update_line .= "PrimEnetBlade='$blade', ";
    $update_line .= "PrimEnetPort='$port' ";
  }
  if ( $Etype =~ m/s/ )  {
    $msg=sprintf("%-20s   Secondary Ethernet:   %10s:%3s:%-2s",
       $ioc,$switch,$blade,$port);
    $update_line .= "SecEnetSwitch='$switch', ";
    $update_line .= "SecEnetBlade='$blade', ";
    $update_line .= "SecEnetPort='$port' ";
  }
  $update_line .= "where ioc_id='$ioc_id'";

  $update_request = qq($update_line);
  eval {
    $sth = $dbh->prepare($update_request);
    $sth->execute();
    $sth->finish();
  };
  if($@) {
    LogUtil::log('error',"update_request: $@");   
    $email_body .= "ERROR: $update_table update execute failed for ioc $ioc\n";
    return -1;
  }
  return 0;
}

#=============================================================================
1;

