#!/usr/bin/perl -w


#
#  Copyright (c) 2004-2005 The University of Chicago, as Operator
#  of Argonne National Laboratory.
#

# network_crawler.pl
#   Reads  accel_sw.log files,and writes resulting network data to db.
#
# Authors: Janet Anderson (jba@aps.anl.gov)
#
#
#

use Mail::Mailer;
use File::stat;
use Cwd;
use Blowfish_PP;

use LogUtil;
use NetworkCrawlerParser;
use NetworkCrawlerDBLayer;

my $status = 0;
my $test_mode = 0;

##########################################################
#  GET SCRIPT ARGUMENTS
##########################################################
$go = 0; # don't run script unless explicitly enabled
$debug_level = "info"; # default
$decryption_key = "bhenKtAdrIdd6Mja";
$replace_all_network_data = 0;

if (!@ARGV) {
    print_usage();
    exit;
}
while (@ARGV) {
    $command = shift @ARGV;
    if ($command =~ /-help/) {
        print_usage();
        exit;
    } elsif  ($command =~ /--all/) {
        $replace_all_network_data = 1;
    } elsif ($command =~ /--debug-level=(.*)/) {
        $debug_level = $1;
    } elsif ($command =~ /--go/) {
        $go = 1;
    } elsif ($command =~ /--key=(.*)/) {
        # Optional 16 byte blowfish decryption key. If given (non null),
        $decryption_key = $1;
    } elsif ($command =~ /--test/) {
        set_test_mode(1);
        $test_mode =1;
    } else {
        print_usage();
        exit;
    }
}
if (!$go ) {
    print "You must at a minimum use --go command line option to run the script!\n";
    exit;
}

##########
#  SETUP
##########

# Can be 'verbose', 'debug', 'info', 'warn', or 'error'
LogUtil::set_log_level($debug_level);

LogUtil::log('info',"Starting ioc network data update script.");   
$email_body = "Starting ioc network data update script.\n\n";

# Set the input and temporary files
$infile            = "/usr/local/iocapps/iocinfo/network/accel_sw.log";
# =======  Debug settings
$infile            = "./accel_sw.log";
$dir               = "/home/phoebus/JBA/irmis/irmisBase/db/src/crawlers/network";
# =======
my $prev_infile       = "$dir/accel_sw.log_save";
my $infileDiff        = "$dir/accel_sw.log_diff";
$infileDiffSorted  = "$dir/accel_sw.log_diff_sorted";

$sb_in=stat($infile);
$infile_date=scalar localtime $sb_in->mtime;

if ( ! -f $prev_infile ) {
  LogUtil::log('info',"$prev_infile not found.");   
  $email_body .= "$prev_infile not found. Replacing ALL network data.\n";
  $replace_all_network_data = 1;
}

################
#  Setup DBLayer
################

# Tell db layer where the home dir is before we do any chdir.
set_home_dir(getcwd());

# it is assumed the db.properties file is encrypted.
set_decryption_key($decryption_key);


#######################################################
#  Start database update or replace of ioc network data
#######################################################

my @network_hits=();

if ($replace_all_network_data) {

  #########################################################################
  # Replace all ioc network data with data from latest accel_sw.log file
  #########################################################################
  LogUtil::log('info',"Replacing ALL network data.");   
  $email_body .= "Replacing ALL network data.\n";   
  $status = zero_out_network_data();
  if ( $status ==  0 ) {
    $status = get_all_network_data($infile,\@network_hits);
    if ( $status ==  0 ) {
      $status = load_new_network_data(\@network_hits);
    }
  }

} else {

  #############################################################################
  # Update ioc network data from accel_sw.log if differs from prev_accel_sw.log
  #############################################################################
  LogUtil::log('info',"Updating modified network data.");   

  $sb_prev_in=stat($prev_infile);
  $prev_infile_date=scalar localtime $sb_prev_in->mtime;

  $status = create_diff_file();
  if ( $status == 1) {
    LogUtil::log('info',"No data updated.");   
    $email_body .= "No data updated.\n";   
  }
  if ( $status == 0) {
    $status = get_modified_network_data($infileDiffSorted,\@network_hits);
    if ( $status ==  0 ) {
      $status = load_new_network_data(\@network_hits);
    }
  }
  unlink $infileDiff if -f $infileDiff ;
  #unlink $infileDiffSorted if -f $infileDiffSorted ;
}

#########################
#  Save network data file
#########################

if ( $status == 0 && !$test_mode ) {
  unlink $prev_infile if -f $prev_infile;
  $status=system("cp -p $infile $prev_infile");
  if ($status != 0) {
    LogUtil::log('error',"Copy of input file failed. $?");   
    $email_body .=  "\nERROR: Cannot save copy of $infile.\n";
  }
}

LogUtil::log('info',"Finished ioc network data update script.");   
$email_body .= "Finished ioc network data update script.\n";   
send_email($email_body);

exit;

#########################################################################
# Create sorted file of differences
#########################################################################
sub create_diff_file {

  if ( ! -f $infile ) {
    LogUtil::log('error',"$infile not found.");   
    $email_body .= "ERROR: $infile not found.\n";
    return -1;
  }
  LogUtil::log('info',"New accel_sw.log file date: $infile_date");   
  LogUtil::log('info',"Saved accel_sw.log file date: $prev_infile_date");   
  $email_body .= "New accel_sw.log file date: $infile_date\n";
  $email_body .= "Saved accel_sw.log file date: $prev_infile_date\n\n";

  unlink $infileDiff if -f $infileDiff;
  system("diff $prev_infile $infile > $infileDiff");
  if ( !-e $infileDiff ) { 
    LogUtil::log('error',"Network data differences file not created.");
    $email_body .= "ERROR: Network data differences file not created.\n";
    return -1;
  }
  if ( -z  $infileDiff ) {
    LogUtil::log('info',"No network data differences found.");
    $email_body .= "No network data differences found.\n";
    return 1;
  }
  unlink $infileDiffSorted;
  if ( system("sort -k 4.1b,4 -k 1,1  $infileDiff > $infileDiffSorted") != 0 ) {
    LogUtil::log('error',"Cannot sort differences file $infileDiff");
    $email_body .= "ERROR: Cannot sort differences file $infileDiff\n";
    return -1;
  }
  return 0;
}

#########################################################################
# Print out application usage text
#########################################################################
sub print_usage {
    print "usage: network_crawler --go [--all] [--key=<16 char key>] [--test] [--help] [--debug-level=[verbose|debug|info]]\n\n";
    print "The network_crawler reads a file of network switch interface data parsing\n";
    print "the data to determine ioc primary and secondary ethernet switch,blade,\n";
    print "and port settings. They are then inserted into the mysql database\n";
    print "using perl DBI.\n";

    print "Log output goes to standard out, and can be adjusted from the default\n";
    print "\"info\" level. Watch out becuase verbose is VERY verbose.\n\n";

    print "You will need to establish the expected schema by running the\n";
    print "create_pv_tables.sql and alter_pv_tables.sql scripts in the rdbCore\n";
    print "ddl subdirectory. Then edit db.properties to set your local database\n";
    print "connection parameters (ie. user/password etc.).\n\n";

    print "The --test option runs the crawler, but does no db changes.\n";
    print "\n";

    print "You must have --go command line option to run the script!\n";
}

